/* ChanServ functions. */

#include "services.h"

static ChannelInfo *chanlists[256];

const char s_ChanServ[] = "ChanServ";

#include "cs-help.c"


static ChannelInfo *cs_findchan(const char *chan);
static void alpha_insert_chan(ChannelInfo *ci);
static ChannelInfo *makechan(const char *chan, const char *pass,
				const char *desc, const char *founder);
static int delchan(ChannelInfo *ci);
static int is_founder(User *user, ChannelInfo *ci);
static int is_identified(User *user, ChannelInfo *ci);
static int get_access(User *user, ChannelInfo *ci);

static void do_help(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_drop(const char *source);
static void do_set(const char *source);
static void do_set_founder(User *u, ChannelInfo *ci, char *param);
static void do_set_password(User *u, ChannelInfo *ci, char *param);
static void do_set_desc(User *u, ChannelInfo *ci, char *param);
static void do_set_topic(User *u, ChannelInfo *ci, char *param);
static void do_set_mlock(User *u, ChannelInfo *ci, char *param);
static void do_set_keeptopic(User *u, ChannelInfo *ci, char *param);
static void do_set_topiclock(User *u, ChannelInfo *ci, char *param);
static void do_set_private(User *u, ChannelInfo *ci, char *param);
static void do_set_secureops(User *u, ChannelInfo *ci, char *param);
static void do_set_restricted(User *u, ChannelInfo *ci, char *param);
static void do_access(const char *source);
static void do_akick(const char *source);
static void do_info(const char *source);
static void do_list(const char *source);
static void do_getpass(const char *source);

/*************************************************************************/

/* Display total number of registered channels and info about each; or, if
 * a specific channel is given, display information about that channel
 * (like /msg ChanServ INFO <channel>).  If count_only != 0, then only
 * display the number of registered channels (the channel parameter is
 * ignored).
 */

void listchans(int count_only, const char *chan)
{
    long count = 0;
    ChannelInfo *ci;
    int i;

    if (count_only) {

	for (i = 33; i < 256; ++i)
	    for (ci = chanlists[i]; ci; ci = ci->next)
		++count;
	printf("%d channels registered.\n", count);

    } else if (chan) {

	struct tm tm;
	NickInfo *ni;
	char *t;

	if (!(ci = cs_findchan(chan))) {
	    printf("Channel %s not registered.\n", chan);
	    return;
	}
	if (ni = findnick(ci->founder))
	    t = ni->last_usermask;
	else
	    t = NULL;
	printf("Information about channel %s:\n", ci->name);
	printf("    Founder    : %s%s%s%s\n",
			ci->founder, t ? " (" : "", t ? t : "", t ? ")" : "");
	printf("    Description: %s\n", ci->desc);
	tm = *localtime(&ci->time_registered);
	printf("    Registered : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ci->last_used);
	printf("    Last used  : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ci->last_topic)
	    printf("    Last topic : %s\n", ci->last_topic);
	printf("    Mode lock  : ");
	if (ci->mlock_on || ci->mlock_key || ci->mlock_limit) {
	    printf("+%s%s%s%s%s%s%s%s",
			(ci->mlock_on & CMODE_I) ? "i" : "",
			(ci->mlock_key         ) ? "k" : "",
			(ci->mlock_limit       ) ? "l" : "",
			(ci->mlock_on & CMODE_M) ? "m" : "",
			(ci->mlock_on & CMODE_N) ? "n" : "",
			(ci->mlock_on & CMODE_P) ? "p" : "",
			(ci->mlock_on & CMODE_S) ? "s" : "",
			(ci->mlock_on & CMODE_T) ? "t" : "");
	}
	if (ci->mlock_off)
	    printf("-%s%s%s%s%s%s%s%s",
			(ci->mlock_off & CMODE_I) ? "i" : "",
			(ci->mlock_off & CMODE_K) ? "k" : "",
			(ci->mlock_off & CMODE_L) ? "l" : "",
			(ci->mlock_off & CMODE_M) ? "m" : "",
			(ci->mlock_off & CMODE_N) ? "n" : "",
			(ci->mlock_off & CMODE_P) ? "p" : "",
			(ci->mlock_off & CMODE_S) ? "s" : "",
			(ci->mlock_off & CMODE_T) ? "t" : "");
	if (ci->mlock_key)
	    printf(" %s", ci->mlock_key);
	if (ci->mlock_limit)
	    printf(" %ld", ci->mlock_limit);
	printf("\n");

    } else {

	for (i = 33; i < 256; ++i) {
	    for (ci = chanlists[i]; ci; ci = ci->next) {
		printf("%-20s  %s\n", ci->name, ci->desc);
		++count;
	    }
	}
	printf("%d channels registered.\n", count);

    }
}

/*************************************************************************/

/* Main ChanServ routine. */

void chanserv(const char *source, char *buf)
{
    char *cmd, *s;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_ChanServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "REGISTER") == 0) {
	do_register(source);

    } else if (stricmp(cmd, "IDENTIFY") == 0) {
	do_identify(source);

    } else if (stricmp(cmd, "DROP") == 0) {
	do_drop(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (stricmp(cmd, "ACCESS") == 0) {
	do_access(source);

    } else if (stricmp(cmd, "AKICK") == 0) {
	do_akick(source);

    } else if (stricmp(cmd, "INFO") == 0) {
	do_info(source);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else if (!is_services_op(source)) {
	notice(s_ChanServ, source,
"Unrecognized command.  Type \"/msg %s HELP\" for help.", s_ChanServ);

    } else if (stricmp(cmd, "GETPASS") == 0) {
	do_getpass(source);

    } else {
	notice(s_ChanServ, source,
"Unrecognized command.  Type \"/msg %s HELP\" for help.", s_ChanServ);

    }

}

/*************************************************************************/

/* Load/save data files. */

void load_cs_dbase(void)
{
    FILE *f = fopen(CHANSERV_DB, "r");
    int i, j, len;
    ChannelInfo *ci;

    if (!f) {
	log_perror("Can't read ChanServ database " CHANSERV_DB);
	return;
    }

    for (i = 33; i < 256; ++i) {

	while (fgetc(f) == 1) {

	    ci = smalloc(sizeof(ChannelInfo));
	    if (1 != fread(ci, sizeof(ChannelInfo), 1, f)) {
		log_perror("Read error on ChanServ database " CHANSERV_DB);
		exit(20);
	    }
	    /* Can't guarantee the file is in a particular order... */
	    alpha_insert_chan(ci);
	    ci->desc = read_string(f, CHANSERV_DB);
	    if (ci->mlock_key)
		ci->mlock_key = read_string(f, CHANSERV_DB);
	    if (ci->last_topic)
		ci->last_topic = read_string(f, CHANSERV_DB);
	    if (ci->accesscount) {
		ChanAccess *access;
		access = smalloc(sizeof(ChanAccess) * ci->accesscount);
		ci->access = access;
		if (ci->accesscount !=
			fread(access, sizeof(ChanAccess), ci->accesscount, f)) {
		    log_perror("Read error on ChanServ database " CHANSERV_DB);
		    exit(20);
		}
		for (j = 0; j < ci->accesscount; ++j, ++access)
		    access->name = read_string(f, CHANSERV_DB);
		j = 0; access = ci->access;
		/* Clear out unused entries */
		while (j < ci->accesscount) {
		    if (access->is_nick < 0) {
			--ci->accesscount;
			free(access->name);
			if (j < ci->accesscount)
			    bcopy(access+1, access, sizeof(*access) * (ci->accesscount - j));
		    } else {
			++j; ++access;
		    }
		}
		if (ci->accesscount)
		    ci->access = srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	    } /* if (ci->accesscount) */
	    if (ci->akickcount) {
		AutoKick *akick;
		akick = smalloc(sizeof(AutoKick) * ci->akickcount);
		ci->akick = akick;
		if (ci->akickcount !=
			fread(akick, sizeof(AutoKick), ci->akickcount, f)) {
		    log_perror("Read error on ChanServ database " CHANSERV_DB);
		    exit(20);
		}
		for (j = 0; j < ci->akickcount; ++j, ++akick) {
		    akick->name = read_string(f, CHANSERV_DB);
		    if (akick->reason)
			akick->reason = read_string(f, CHANSERV_DB);
		}
		j = 0; akick = ci->akick;
		while (j < ci->akickcount) {
		    if (akick->is_nick < 0) {
			--ci->akickcount;
			free(akick->name);
			if (akick->reason)
			    free(akick->reason);
			if (j < ci->akickcount)
			    bcopy(akick+1, akick, sizeof(*akick) * (ci->akickcount - j));
		    } else {
			++j; ++akick;
		    }
		}
		if (ci->akickcount)
		    ci->akick = srealloc(ci->akick, sizeof(AutoKick) * ci->akickcount);
	    } /* if (ci->akickcount) */

	} /* while (fgetc(f) == 1) */

    } /* for (i) */

    fclose(f);
}

/*************************************************************************/

void save_cs_dbase(void)
{
    FILE *f;
    int i, j, len;
    ChannelInfo *ci;

    remove(CHANSERV_DB ".save");
    rename(CHANSERV_DB, CHANSERV_DB ".save");
    f = fopen(CHANSERV_DB, "w");
    if (!f) {
	log_perror("Can't write to ChanServ database " CHANSERV_DB);
	return;
    }
    for (i = 33; i < 256; ++i) {
	for (ci = chanlists[i]; ci; ci = ci->next) {
	    fputc(1, f);
	    if (1 != fwrite(ci, sizeof(ChannelInfo), 1, f)) {
		log_perror("Write error on ChanServ database " CHANSERV_DB);
		wallops("Write error");
		exit(1);
	    }
	    write_string(ci->desc, f, CHANSERV_DB);
	    if (ci->mlock_key)
		write_string(ci->mlock_key, f, CHANSERV_DB);
	    if (ci->last_topic)
		write_string(ci->last_topic, f, CHANSERV_DB);
	    if (ci->accesscount) {
		ChanAccess *access = ci->access;
		if (ci->accesscount !=
			fwrite(access, sizeof(ChanAccess), ci->accesscount, f)) {
		    log_perror("Write error on ChanServ database " CHANSERV_DB);
		    wallops("Write error");
		    exit(1);
		}
		for (j = 0; j < ci->accesscount; ++j, ++access)
		    write_string(access->name, f, CHANSERV_DB);
	    }
	    if (ci->akickcount) {
		AutoKick *akick = ci->akick;
		if (ci->akickcount !=
			fwrite(akick, sizeof(AutoKick), ci->akickcount, f)) {
		    log_perror("Write error on ChanServ database " CHANSERV_DB);
		    wallops("Write error");
		    exit(1);
		}
		for (j = 0; j < ci->akickcount; ++j, ++akick) {
		    write_string(akick->name, f, CHANSERV_DB);
		    if (akick->reason)
			write_string(akick->reason, f, CHANSERV_DB);
		}
	    }
	} /* for (chanlists[i]) */
	fputc(0, f);
    } /* for (i) */
    fclose(f);
    remove(CHANSERV_DB ".save");
}

/*************************************************************************/

/* Check the current modes on a channel; if they conflict with a mode lock,
 * fix them. */

void check_modes(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci = cs_findchan(chan);
    char newmodes[32], *newkey = NULL;
    long newlimit = 0;
    char *end = newmodes;
    int modes;

    if (!c || !ci)
	return;

    *end++ = '+';
    modes = ~c->mode & ci->mlock_on;
    if (modes & CMODE_I) {
	*end++ = 'i';
	c->mode |= CMODE_I;
    }
    if (modes & CMODE_M) {
	*end++ = 'm';
	c->mode |= CMODE_M;
    }
    if (modes & CMODE_N) {
	*end++ = 'n';
	c->mode |= CMODE_N;
    }
    if (modes & CMODE_P) {
	*end++ = 'p';
	c->mode |= CMODE_P;
    }
    if (modes & CMODE_S) {
	*end++ = 's';
	c->mode |= CMODE_S;
    }
    if (modes & CMODE_T) {
	*end++ = 't';
	c->mode |= CMODE_T;
    }

    if (!c->key && ci->mlock_key) {
	*end++ = 'k';
	newkey = ci->mlock_key;
	c->key = sstrdup(newkey);
    }
    if (!c->limit && ci->mlock_limit) {
	*end++ = 'l';
	newlimit = ci->mlock_limit;
	c->limit = newlimit;
    }

    if (end[-1] == '+')
	--end;

    *end++ = '-';
    modes = c->mode & ci->mlock_off;
    if (modes & CMODE_I) {
	*end++ = 'i';
	c->mode &= ~CMODE_I;
    }
    if (modes & CMODE_M) {
	*end++ = 'm';
	c->mode &= ~CMODE_M;
    }
    if (modes & CMODE_N) {
	*end++ = 'n';
	c->mode &= ~CMODE_N;
    }
    if (modes & CMODE_P) {
	*end++ = 'p';
	c->mode &= ~CMODE_P;
    }
    if (modes & CMODE_S) {
	*end++ = 's';
	c->mode &= ~CMODE_S;
    }
    if (modes & CMODE_T) {
	*end++ = 't';
	c->mode &= ~CMODE_T;
    }

    if (c->key && (ci->mlock_off & CMODE_K)) {
	*end++ = 'k';
	newkey = sstrdup(c->key);
	free(c->key);
	c->key = NULL;
    }
    if (c->limit && (ci->mlock_off & CMODE_L)) {
	*end++ = 'l';
	c->limit = 0;
    }

    if (end[-1] == '-')
	--end;

    if (end == newmodes)
	return;
    *end = 0;
    send_cmd(s_ChanServ, "MODE %s %s %s %d", c->name,
				newmodes, newkey ? newkey : "", newlimit);
    if (newkey && !c->key)
	free(newkey);
}

/*************************************************************************/

/* Check whether a user is allowed to be opped on a channel; if they
 * aren't, deop them.  If serverop is 1, the +o was done by a server.
 * Return 1 if the user is allowed to be opped, 0 otherwise. */

int check_valid_op(User *user, const char *chan, int serverop)
{
    ChannelInfo *ci = cs_findchan(chan);

    if (!ci)
	return 1;

    if (serverop && get_access(user, ci) < 1) {
	notice(s_ChanServ, user->nick,
		"This channel has been registered with %s.", s_ChanServ);
/*notice(SERVER_NAME, "Alcan", "check_valid_op: sending: MODE %s -o %s",
chan, user->nick);*/
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
	return 0;
    }

    if (get_access(user, ci) < (ci->flags & CI_SECUREOPS) ? 1 : 0) {
	notice(s_ChanServ, user->nick,
		"You are not allowed chanop status on channel %s.", chan);
/*notice(SERVER_NAME, "Alcan", "check_valid_op: sending: MODE %s -o %s",
chan, user->nick);*/
	send_cmd(s_ChanServ, "MODE %s -o %s", chan, user->nick);
	return 0;
    }

    return 1;
}

/*************************************************************************/

/* Check whether a user should be opped on a channel, and if so, do it.
 * Return 1 if the user was opped, 0 otherwise.  (Updates the channel's
 * last used time if the user was opped.) */

int check_should_op(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);

    if (!ci)
	return 0;

    if (get_access(user, ci) >= 5) {
/*notice(SERVER_NAME, "Alcan", "check_should_op: sending: MODE %s +o %s",
chan, user->nick);*/
	send_cmd(s_ChanServ, "MODE %s +o %s", chan, user->nick);
	ci->last_used = time(NULL);
	return 1;
    }

    return 0;
}

/*************************************************************************/

/* Check whether a user is permitted to be on a channel.  If so, return 0;
 * else, kickban the user with an appropriate message (could be either
 * AKICK or restricted access). */

int check_kick(User *user, const char *chan)
{
    ChannelInfo *ci = cs_findchan(chan);
    AutoKick *akick;
    int i, bad;
    NickInfo *ni;
    char *av[3];

    if (!ci)
	return 0;

    ni = findnick(user->nick);
    if (!(ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)))
	ni = NULL;

    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i) {
	if ((akick->is_nick > 0 && ni && stricmp(akick->name, user->nick) == 0) ||
		(!akick->is_nick && match_usermask(akick->name, user))) {
	    av[0] = sstrdup(chan);
	    av[1] = "+b";
	    av[2] = akick->is_nick ? create_mask(user) : akick->name;
	    send_cmd(s_ChanServ, "MODE %s +b %s", chan, av[2]);
	    do_cmode(s_ChanServ, 3, av);
	    send_cmd(s_ChanServ,
			"KICK %s %s :%s", chan, user->nick,
			akick->reason ? akick->reason :
				"User has been banned from the channel");
	    free(av[0]);
	    if (akick->is_nick)
		free(av[2]);
	    return 1;
	}
    }

    if (!(ci->flags & CI_RESTRICTED))
	return 0;

    if (get_access(user, ci) >= ((ci->flags&CI_SECUREOPS) ? 1 : 0)) {
	av[0] = sstrdup(chan);
	av[1] = "+b";
	av[2] = create_mask(user);
	send_cmd(s_ChanServ, "MODE %s +b %s", chan, av[2]);
	do_cmode(s_ChanServ, 3, av);
	send_cmd(s_ChanServ,
		"KICK %s %s :%s", chan, user->nick,
		"You are not permitted to be on this channel.");
	free(av[0]);
	free(av[2]);
	return 1;
    }

    return 0;
}

/*************************************************************************/

/* Record the current channel topic in the ChannelInfo structure. */

void record_topic(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci = cs_findchan(chan);

    if (!c || !ci)
	return;
    if (ci->last_topic)
	free(ci->last_topic);
    if (c->topic)
	ci->last_topic = sstrdup(c->topic);
    else
	ci->last_topic = NULL;
    strscpy(ci->last_topic_setter, c->topic_setter, NICKMAX);
    ci->last_topic_time = c->topic_time;
}

/*************************************************************************/

/* Restore the topic in a channel when it's created, if we should. */

void restore_topic(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci = cs_findchan(chan);

    if (!c || !ci || !(ci->flags & CI_KEEPTOPIC))
	return;
    if (c->topic)
	free(c->topic);
    if (ci->last_topic) {
	c->topic = sstrdup(ci->last_topic);
	strscpy(c->topic_setter, ci->last_topic_setter, NICKMAX);
	c->topic_time = ci->last_topic_time;
    } else
	c->topic = NULL;
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s", chan,
		c->topic_setter, c->topic_time, c->topic ? c->topic : "");
}

/*************************************************************************/

/* See if the topic is locked on the given channel, and return 1 (and fix
 * the topic) if so. */

int check_topiclock(const char *chan)
{
    Channel *c = findchan(chan);
    ChannelInfo *ci = cs_findchan(chan);

    if (!c || !ci || !(ci->flags & CI_TOPICLOCK))
	return 0;
    if (c->topic)
	free(c->topic);
    if (ci->last_topic)
	c->topic = sstrdup(ci->last_topic);
    else
	c->topic = NULL;
    strscpy(c->topic_setter, ci->last_topic_setter, NICKMAX);
    c->topic_time = ci->last_topic_time;
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s", chan,
		c->topic_setter, c->topic_time, c->topic ? c->topic : "");
}

/*************************************************************************/

/* Remove all channels which have expired. */

void expire_chans()
{
    ChannelInfo *ci;
    int i;
    const time_t expire_time = 14*24*60*60;
    time_t now = time(NULL);

    for (i = 33; i < 256; ++i) {
	for (ci = chanlists[i]; ci; ci = ci->next) {
	    if (now - ci->last_used >= expire_time)
		delchan(ci);
	}
    }
}

/*************************************************************************/
/*********************** ChanServ private routines ***********************/
/*************************************************************************/

/* Return the ChannelInfo structure for the given channel, or NULL if the
 * channel isn't registered. */

static ChannelInfo *cs_findchan(const char *chan)
{
    ChannelInfo *ci;

    for (ci = chanlists[tolower(chan[1])]; ci; ci = ci->next) {
	if (stricmp(ci->name, chan) == 0)
	    return ci;
    }
    return NULL;
}

/*************************************************************************/

/* Insert a channel alphabetically into the database. */

static void alpha_insert_chan(ChannelInfo *ci)
{
    ChannelInfo *ci2, *ci3;
    char *chan = ci->name;

    for (ci3 = NULL, ci2 = chanlists[tolower(chan[1])];
			ci2 && stricmp(ci2->name, chan) < 0;
			ci3 = ci2, ci2 = ci2->next)
	;
    ci->prev = ci3;
    ci->next = ci2;
    if (!ci3)
	chanlists[tolower(chan[1])] = ci;
    else
	ci3->next = ci;
    if (ci2)
	ci2->prev = ci;
}

/*************************************************************************/

/* Add a channel to the database.  Returns a pointer to the new ChannelInfo
 * structure if the channel was successfully registered, NULL otherwise.
 * Assumes channel does not already exist. */

static ChannelInfo *makechan(const char *chan, const char *pass,
				const char *desc, const char *founder)
{
    ChannelInfo *ci;
    Channel *c;
    char *s, *end;

    if (!(c = findchan(chan))) {
	log("%s: makechan: channel %s not found", s_ChanServ, chan);
	return NULL;
    }
    ci = scalloc(sizeof(ChannelInfo), 1);
    ci->last_used = ci->time_registered = time(NULL);
    strscpy(ci->name, chan, CHANMAX);
    strscpy(ci->founder, founder, NICKMAX);
    strscpy(ci->founderpass, pass, PASSMAX);
    ci->desc = sstrdup(desc);
    if (c->topic) {
	ci->last_topic = sstrdup(c->topic);
	strscpy(ci->last_topic_setter, c->topic_setter, NICKMAX);
	ci->last_topic_time = c->topic_time;
    }
    ci->flags = CI_KEEPTOPIC;
    alpha_insert_chan(ci);
    return ci;
}

/*************************************************************************/

/* Remove a channel from the ChanServ database.  Return 1 on success, 0
 * otherwise. */

static int delchan(ChannelInfo *ci)
{
    int i;

    if (ci->next)
	ci->next->prev = ci->prev;
    if (ci->prev)
	ci->prev->next = ci->next;
    else
	chanlists[tolower(ci->name[1])] = ci->next;
    if (ci->mlock_key)
	free(ci->mlock_key);
    if (ci->last_topic)
	free(ci->last_topic);
    for (i = 0; i < ci->accesscount; ++i)
	free(ci->access[i].name);
    free(ci->access);
    for (i = 0; i < ci->akickcount; ++i) {
	free(ci->akick[i].name);
	if (ci->akick[i].reason)
	    free(ci->akick[i].reason);
    }
    free(ci->akick);
    return 1;
}

/*************************************************************************/

/* Does the given user have founder access to the channel? */

static int is_founder(User *user, ChannelInfo *ci)
{
    struct u_chaninfolist *c;
    NickInfo *ni;

    if ((ni = findnick(user->nick)) &&
		(ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)) &&
		stricmp(user->nick, ci->founder) == 0)
	return 1;
    for (c = user->founder_chans; c; c = c->next) {
	if (c->chan == ci)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Has the given user password-identified as founder for the channel? */

static int is_identified(User *user, ChannelInfo *ci)
{
    struct u_chaninfolist *c;

    for (c = user->founder_chans; c; c = c->next) {
	if (c->chan == ci)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Return the access level the given user has on the channel.  If the
 * channel doesn't exist or the user isn't on the access list, return 0. */

static int get_access(User *user, ChannelInfo *ci)
{
    NickInfo *ni = findnick(user->nick);
    ChanAccess *access;
    int i;

    if (!ci)
	return 0;
    if (is_founder(user, ci) || is_identified(user, ci))
	return 10000;
    for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i) {
	if (		(access->is_nick > 0 &&
			 ni &&
			 (ni->flags & (NI_IDENTIFIED | NI_RECOGNIZED)) &&
			 stricmp(user->nick, access->name) == 0
			) ||
			(!access->is_nick &&
			 match_usermask(access->name, user)
			)
	   )
	    return access->level;
    }
    return 0;
}

/*************************************************************************/
/*********************** ChanServ command routines ***********************/
/*************************************************************************/

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, " ");

    if (!cmd)
	notice_list(s_ChanServ, source, cs_help);

    else if (stricmp(cmd, "REGISTER") == 0)
	notice_list(s_ChanServ, source, register_help);

    else if (stricmp(cmd, "IDENTIFY") == 0)
	notice_list(s_ChanServ, source, identify_help);

    else if (stricmp(cmd, "DROP") == 0) {
	if (is_oper(source))
	    notice_list(s_ChanServ, source, oper_drop_help);
	else
	    notice_list(s_ChanServ, source, drop_help);

    } else if (stricmp(cmd, "SET") == 0) {

	cmd = strtok(NULL, " ");

	if (!cmd)
	    notice_list(s_ChanServ, source, set_help);

	else if (stricmp(cmd, "FOUNDER") == 0)
	    notice_list(s_ChanServ, source, set_founder_help);

	else if (stricmp(cmd, "PASSWORD") == 0)
	    notice_list(s_ChanServ, source, set_pass_help);

	else if (stricmp(cmd, "DESC") == 0)
	    notice_list(s_ChanServ, source, set_desc_help);

	else if (stricmp(cmd, "TOPIC") == 0)
	    notice_list(s_ChanServ, source, set_topic_help);

	else if (stricmp(cmd, "MLOCK") == 0)
	    notice_list(s_ChanServ, source, set_mlock_help);

	else if (stricmp(cmd, "KEEPTOPIC") == 0)
	    notice_list(s_ChanServ, source, set_keeptopic_help);

	else if (stricmp(cmd, "TOPICLOCK") == 0)
	    notice_list(s_ChanServ, source, set_topiclock_help);

	else if (stricmp(cmd, "PRIVATE") == 0)
	    notice_list(s_ChanServ, source, set_private_help);

	else if (stricmp(cmd, "SECUREOPS") == 0)
	    notice_list(s_ChanServ, source, set_secureops_help);

	else if (stricmp(cmd, "RESTRICTED") == 0)
	    notice_list(s_ChanServ, source, set_restricted_help);

	else
	    notice(s_ChanServ, source,
		"No help available for \2SET %s\2", strupper(cmd));

    } else if (stricmp(cmd, "ACCESS") == 0) {
	cmd = strtok(NULL, " ");
	if (cmd && stricmp(cmd, "LEVELS") == 0)
	    notice_list(s_ChanServ, source, levels_help);
	else
	    notice_list(s_ChanServ, source, access_help);

    } else if (stricmp(cmd, "AKICK") == 0)
	notice_list(s_ChanServ, source, akick_help);

    else if (stricmp(cmd, "INFO") == 0)
	notice_list(s_ChanServ, source, info_help);

    else if (stricmp(cmd, "LIST") == 0)
	notice_list(s_ChanServ, source, list_help);

    else if (stricmp(cmd, "INVITE") == 0)
	notice_list(s_ChanServ, source, invite_help);

    else if (stricmp(cmd, "UNBAN") == 0)
	notice_list(s_ChanServ, source, unban_help);

    else if (!is_services_op(source))
	notice(s_ChanServ, source,
			"No help available for \2%s\2", strupper(cmd));

    else if (stricmp(cmd, "GETPASS") == 0)
	notice_list(s_ChanServ, source, getpass_help);

    else
	notice(s_ChanServ, source,
		"No help available for \2%s\2.", strupper(cmd));
}

/*************************************************************************/

static void do_register(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    char *desc = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    ChannelInfo *ci;
    User *u = finduser(source);
    struct u_chaninfolist *c;

    if (!desc) {

	notice(s_ChanServ, source,
		"Syntax: \2REGISTER \37channel\37 \37password\37 \37description\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP REGISTER\2 for more information.", s_ChanServ);

    } else if (*chan == '&') {

	notice(s_ChanServ, source, "Local channels cannot be registered.");

    } else if (chan[1] == 0) {

	notice(s_ChanServ, source, "%s is not a valid channel name.", chan);

    } else if (!ni) {

	notice(s_ChanServ, source, "Your must register your nickname first.");
	notice(s_ChanServ, source,
		"\2/msg %s HELP\2 for information on registering nicknames.",
		s_NickServ);

    } else if (cs_findchan(chan)) {

	notice(s_ChanServ, source, "Channel %s already registered!", chan);

    } else if (!u) {

	log("%s: Attempt to register channel %s from nonexistent nick %s",
						s_ChanServ, chan, source);
	notice(s_ChanServ, source, "Sorry, registration failed.");

    } else if (!is_chanop(source, chan)) {

	notice(s_ChanServ, source,
		"You must be a channel operator to register the channel.");

    } else {

	if (ci = makechan(chan, pass, desc, source)) {
	    notice(s_ChanServ, source,
			"Channel %s registered under your nickname: %s",
			chan, source);
	    notice(s_ChanServ, source,
			"Your channel password is \2%s\2 - remember it for "
			"later use.", ci->founderpass);
	} else {
	    notice(s_ChanServ, source, "Sorry, registration failed.");
	}
	c = smalloc(sizeof(*c));
	c->next = u->founder_chans;
	c->prev = NULL;
	if (u->founder_chans)
	    u->founder_chans->prev = c;
	u->founder_chans = c;
	c->chan = ci;

    }
}

/*************************************************************************/

static void do_identify(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    ChannelInfo *ci;
    struct u_chaninfolist *c;
    User *u = finduser(source);;

    if (!pass) {

	notice(s_ChanServ, source,
		"Syntax: \2IDENTIFY %s \37password\37\2",
		chan ? chan : "\37channel\37");
	notice(s_ChanServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!u) {

	log("%s: IDENTIFY from nonexistent nick %s for %s",
						s_ChanServ, source, chan);
	notice(s_ChanServ, source, "Sorry, identification failed.");

    } else {

	if (strcmp(pass, ci->founderpass) == 0) {
	    if (!is_identified(u, ci)) {
		c = smalloc(sizeof(*c));
		c->next = u->founder_chans;
		c->prev = NULL;
		if (u->founder_chans)
		    u->founder_chans->prev = c;
		u->founder_chans = c;
		c->chan = ci;
	    }
	    notice(s_ChanServ, source,
		"Password accepted - you now have founder-level access to %s.",
		chan);
	} else {
	    log("%s: Failed IDENTIFY for %s by %s!%s@%s",
			s_ChanServ, chan, source, u->username, u->host);
	    notice(s_ChanServ, source, "Password incorrect.");
	}

    }
}

/*************************************************************************/

static void do_drop(const char *source)
{
    char *chan = strtok(NULL, " ");
    ChannelInfo *ci;
    User *u;

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2DROP \37channel\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP DROP\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!is_services_op(source) &&
		(!(u = finduser(source)) || !is_identified(u, ci))) {

	notice(s_ChanServ, source,
		"Password authentication required for that command.");
	notice(s_ChanServ, source,
		"Retry after typing \2/msg %s IDENTIFY %s <password>.",
		s_ChanServ, chan);

    } else {

	delchan(ci);
	notice(s_ChanServ, source, "Channel %s has been dropped.", chan);

    }
}

/*************************************************************************/

/* Main SET routine.  Calls other routines as follows:
 *	do_set_command(User *command-sender, ChannelInfo *ci, char *param);
 * Additional parameters can be retrieved using strtok(NULL, toks).
 */
static void do_set(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *param;
    ChannelInfo *ci;
    User *u;
    int i;

    if (cmd) {
	if (stricmp(cmd, "DESC") == 0 || stricmp(cmd, "TOPIC") == 0)
	    param = strtok(NULL, "");
	else
	    param = strtok(NULL, " ");
    } else
	param = NULL;

    if (!param) {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s \37option\37 \37parameters\37\2",
		chan ? chan : "\37channel\37");
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!(u = finduser(source)) || !is_identified(u, ci)) {

	if (is_founder(u, ci)) {
	    notice(s_ChanServ, source,
		"Password authentication required for that command.");
	    notice(s_ChanServ, source,
		"Retry after typing \2/msg ChanServ IDENTIFY %s \37password\37.",
		chan);
	} else {
	    notice(s_ChanServ, source, "Access denied.");
	}

    } else if (stricmp(cmd, "FOUNDER") == 0) {

	do_set_founder(u, ci, param);

    } else if (stricmp(cmd, "PASSWORD") == 0) {

	do_set_password(u, ci, param);

    } else if (stricmp(cmd, "DESC") == 0) {

	do_set_desc(u, ci, param);

    } else if (stricmp(cmd, "TOPIC") == 0) {

	do_set_topic(u, ci, param);

    } else if (stricmp(cmd, "MLOCK") == 0) {

	do_set_mlock(u, ci, param);

    } else if (stricmp(cmd, "KEEPTOPIC") == 0) {

	do_set_keeptopic(u, ci, param);

    } else if (stricmp(cmd, "TOPICLOCK") == 0) {

	do_set_topiclock(u, ci, param);

    } else if (stricmp(cmd, "PRIVATE") == 0) {

	do_set_private(u, ci, param);

    } else if (stricmp(cmd, "SECUREOPS") == 0) {

	do_set_secureops(u, ci, param);

    } else if (stricmp(cmd, "RESTRICTED") == 0) {

	do_set_restricted(u, ci, param);

    } else {

	notice(s_ChanServ, source,
			"Unknown SET option \2%s\2.", strupper(cmd));

    }
}

/*************************************************************************/

static void do_set_founder(User *u, ChannelInfo *ci, char *param)
{
    NickInfo *ni = findnick(param);

    if (!ni) {
	notice(s_ChanServ, u->nick, "%s isn't a registered nickname.", param);
	return;
    }
    strscpy(ci->founder, param, NICKMAX);
    notice(s_ChanServ, u->nick,
		"Founder of %s changed to \2%s\2.", ci->name, ci->founder);
}

/*************************************************************************/

static void do_set_password(User *u, ChannelInfo *ci, char *param)
{
    strscpy(ci->founderpass, param, PASSMAX);
    notice(s_ChanServ, u->nick,
		"%s password changed to \2%s\2.", ci->name, ci->founderpass);
}

/*************************************************************************/

static void do_set_desc(User *u, ChannelInfo *ci, char *param)
{
    free(ci->desc);
    ci->desc = sstrdup(param);
    notice(s_ChanServ, u->nick,
		"Description of %s changed to \2%s\2.", ci->name, param);
}

/*************************************************************************/

static void do_set_topic(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;
    Channel *c = findchan(chan);

    if (!c) {
	log("%s: SET TOPIC for %s from %s: channel not found!",
		s_ChanServ, chan, source);
	notice(s_ChanServ, source, "Sorry, can't set topic.");
	return;
    }
    if (ci->last_topic)
	free(ci->last_topic);
    if (*param)
	ci->last_topic = sstrdup(param);
    else
	ci->last_topic = NULL;
    if (c->topic)
	free(c->topic);
    if (*param)
	c->topic = sstrdup(param);
    else
	c->topic = NULL;
    strscpy(ci->last_topic_setter, source, NICKMAX);
    strscpy(c->topic_setter, source, NICKMAX);
    ci->last_topic_time = c->topic_time = time(NULL);
notice(s_ChanServ, "Alcan", "TOPIC %s %s %lu :%s",
chan, source, c->topic_time, param);
    send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s",
		chan, source, c->topic_time, param);
}

/*************************************************************************/

static void do_set_mlock(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;
    char *s, modebuf[32], *end;
    int add = -1;	/* 1 if adding, 0 if deleting, -1 if neither */

    ci->mlock_on = ci->mlock_off = ci->mlock_limit = 0;
    if (ci->mlock_key)
	free(ci->mlock_key);

    while (*param) {
	if (*param != '+' && *param != '-' && add < 0) {
	    ++param;
	    continue;
	}
	switch (tolower(*param++)) {
	case '+':
	    add = 1; break;
	case '-':
	    add = 0; break;
	case 'i':
	    if (add)
		ci->mlock_on |= CMODE_I;
	    else
		ci->mlock_off |= CMODE_I;
	    break;
	case 'k':
	    if (add) {
		if (!(s = strtok(NULL, " ")))
		    notice(s_ChanServ, source,
				"Parameter required for MLOCK +k.");
		ci->mlock_key = sstrdup(s);
	    } else
		ci->mlock_off |= CMODE_K;
	    break;
	case 'l':
	    if (add) {
		if (!(s = strtok(NULL, " ")))
		    notice(s_ChanServ, source,
				"Parameter required for MLOCK +l.");
		if (atol(s) <= 0)
		    notice(s_ChanServ, source,
				"Parameter for MLOCK +l must be a positive number.");
		ci->mlock_limit = atol(s);
	    } else
		ci->mlock_off |= CMODE_L;
	    break;
	case 'm':
	    if (add)
		ci->mlock_on |= CMODE_M;
	    else
		ci->mlock_off |= CMODE_M;
	    break;
	case 'n':
	    if (add)
		ci->mlock_on |= CMODE_N;
	    else
		ci->mlock_off |= CMODE_N;
	    break;
	case 'p':
	    if (add)
		ci->mlock_on |= CMODE_P;
	    else
		ci->mlock_off |= CMODE_P;
	    break;
	case 's':
	    if (add)
		ci->mlock_on |= CMODE_S;
	    else
		ci->mlock_off |= CMODE_S;
	    break;
	case 't':
	    if (add)
		ci->mlock_on |= CMODE_T;
	    else
		ci->mlock_off |= CMODE_T;
	    break;
	} /* switch */
    } /* while (*param) */

    end = modebuf;
    *end = 0;
    if (ci->mlock_on)
	end += sprintf(end, "+%s%s%s%s%s%s%s%s",
				(ci->mlock_on & CMODE_I) ? "i" : "",
				(ci->mlock_key         ) ? "k" : "",
				(ci->mlock_limit       ) ? "l" : "",
				(ci->mlock_on & CMODE_M) ? "m" : "",
				(ci->mlock_on & CMODE_N) ? "n" : "",
				(ci->mlock_on & CMODE_P) ? "p" : "",
				(ci->mlock_on & CMODE_S) ? "s" : "",
				(ci->mlock_on & CMODE_T) ? "t" : "");
    if (ci->mlock_off)
	end += sprintf(end, "-%s%s%s%s%s%s%s%s",
				(ci->mlock_off & CMODE_I) ? "i" : "",
				(ci->mlock_off & CMODE_K) ? "k" : "",
				(ci->mlock_off & CMODE_L) ? "l" : "",
				(ci->mlock_off & CMODE_M) ? "m" : "",
				(ci->mlock_off & CMODE_N) ? "n" : "",
				(ci->mlock_off & CMODE_P) ? "p" : "",
				(ci->mlock_off & CMODE_S) ? "s" : "",
				(ci->mlock_off & CMODE_T) ? "t" : "");
    notice(s_ChanServ, source,
		"Mode lock on channel %s changed to \2%s\2.", chan, modebuf);
    check_modes(chan);
}

/*************************************************************************/

static void do_set_keeptopic(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp(param, "ON") == 0) {

	ci->flags |= CI_KEEPTOPIC;
	notice(s_ChanServ, source,
		"Topic retention option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ci->flags &= ~CI_KEEPTOPIC;
	notice(s_ChanServ, source,
		"Topic retention option is now \2OFF\2.");

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s KEEPTOPIC {ON|OFF}\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET KEEPTOPIC\2 for more information.",
		s_ChanServ);
    }
}

/*************************************************************************/

static void do_set_topiclock(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp(param, "ON") == 0) {

	ci->flags |= CI_TOPICLOCK;
	notice(s_ChanServ, source,
		"Topic lock option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ci->flags &= ~CI_TOPICLOCK;
	notice(s_ChanServ, source,
		"Topic lock option is now \2OFF\2.");

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s TOPICLOCK {ON|OFF}\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET TOPICLOCK\2 for more information.",
		s_ChanServ);
    }
}

/*************************************************************************/

static void do_set_private(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp(param, "ON") == 0) {

	ci->flags |= CI_PRIVATE;
	notice(s_ChanServ, source,
		"Private option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ci->flags &= ~CI_PRIVATE;
	notice(s_ChanServ, source,
		"Private option is now \2OFF\2.");

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s PRIVATE {ON|OFF}\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET PRIVATE\2 for more information.",
		s_ChanServ);
    }
}

/*************************************************************************/

static void do_set_secureops(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp(param, "ON") == 0) {

	ci->flags |= CI_SECUREOPS;
	notice(s_ChanServ, source,
		"Secure ops option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ci->flags &= ~CI_SECUREOPS;
	notice(s_ChanServ, source,
		"Secure ops option is now \2OFF\2.");

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s SECUREOPS {ON|OFF}\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET SECUREOPS\2 for more information.",
		s_ChanServ);
    }
}

/*************************************************************************/

static void do_set_restricted(User *u, ChannelInfo *ci, char *param)
{
    char *source = u->nick, *chan = ci->name;

    if (stricmp(param, "ON") == 0) {

	ci->flags |= CI_RESTRICTED;
	notice(s_ChanServ, source,
		"Restricted access option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ci->flags &= ~CI_RESTRICTED;
	notice(s_ChanServ, source,
		"Restricted access option is now \2OFF\2.");

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2SET %s RESTRICTED {ON|OFF}\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP SET RESTRICTED\2 for more information.",
		s_ChanServ);
    }
}

/*************************************************************************/

static void do_access(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    char *s    = strtok(NULL, " ");
    ChannelInfo *ci;
    NickInfo *ni;
    User *u;
    int level, ulev, i;
    ChanAccess *access;
    char *nick, *user, *host;

    if (!cmd || ((stricmp(cmd,"LIST")==0) ? 0 :
			(stricmp(cmd,"DEL")==0) ? (!mask || !!s) : !s)) {
	notice(s_ChanServ, source,
		"Syntax: \2ACCESS %s {ADD|DEL|LIST} [\37user\37 [\37level\37]]\2",
		chan ? chan : "\37channel\37");
	notice(s_ChanServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!(u = finduser(source)) || ((ulev = get_access(u, ci)) <= 0) ||
						(level = atoi(s)) >= ulev) {
	notice(s_ChanServ, source, "Access denied.");

    } else if ((level = atoi(s)) == 0 && stricmp(cmd, "ADD") == 0) {

	notice(s_ChanServ, source, "Access level must be non-zero.");

    } else if (stricmp(cmd, "ADD") == 0) {

	ni = findnick(mask);
	if (!ni) {
	    split_usermask(mask, &nick, &user, &host);
	    s = smalloc(strlen(nick)+strlen(user)+strlen(host)+3);
	    sprintf(s, "%s!%s@%s", nick, user, host);
	    free(nick);
	    free(user);
	    free(host);
	}
	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i) {
	    if ((access->is_nick ? stricmp(access->name,mask) :
					strcmp(access->name,mask)) == 0) {
		if (access->level == level) {
		    notice(s_ChanServ, source,
			"Access level for \2%s\2 on %s unchanged from \2%d\2.",
			access->name, chan, level);
		    return;
		}
		access->level = level;
		notice(s_ChanServ, source,
			"Access level for \2%s\2 on %s changed to \2%d\2.",
			access->name, chan, level);
		return;
	    }
	}
	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i) {
	    if (access->is_nick < 0)
		break;
	}
	if (i == ci->accesscount) {
	    ++ci->accesscount;
	    ci->access =
		srealloc(ci->access, sizeof(ChanAccess) * ci->accesscount);
	    access = &ci->access[ci->accesscount-1];
	}
	access->name = ni ? sstrdup(ni->nick) : s;
	access->is_nick = (ni != NULL);
	access->level = level;
	notice(s_ChanServ, source,
		"\2%s\2 added to %s access list at level \2%d\2.",
		access->name, chan, level);

    } else if (stricmp(cmd, "DEL") == 0) {

	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn(mask, "1234567890") == strlen(mask) &&
				(i = atoi(mask)) > 0 && i <= ci->accesscount) {
	    --i;
	    access = &ci->access[i];
	} else {
	    ni = findnick(mask);
	    if (!ni) {
		split_usermask(mask, &nick, &user, &host);
		mask = smalloc(strlen(nick)+strlen(user)+strlen(host)+3);
		sprintf(mask, "%s!%s@%s", nick, user, host);
		free(nick);
		free(user);
		free(host);
	    }
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (access = ci->access, i = 0; i < ci->accesscount;
							++access, ++i) {
		if (strcmp(access->name, mask) == 0)
		    break;
	    }
	    if (i == ci->accesscount) {
		for (access = ci->access, i = 0; i < ci->accesscount;
							++access, ++i) {
		    if (stricmp(access->name, mask) == 0)
			break;
		}
	    }
	    if (i == ci->accesscount) {
		notice(s_ChanServ, source,
			"\2%s\2 not found on %s access list.", mask, chan);
		if (!ni)
		    free(mask);
		return;
	    }
	}
	notice(s_ChanServ, source,
		"\2%s\2 deleted from %s access list.", access->name, chan);
	free(access->name);
	access->is_nick = -1;
	if (!ni)
	    free(mask);

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_ChanServ, source, "Access list for %s:", chan);
	notice(s_ChanServ, source, "  Num   Lev  Mask");
	for (access = ci->access, i = 0; i < ci->accesscount; ++access, ++i) {
	    if (access->is_nick < 0)
		continue;
	    if ((mask && !match_wild(mask, access->name)) ||
					(level && level != access->level))
		continue;
	    if (ni = findnick(access->name))
		s = ni->last_usermask;
	    else
		s = NULL;
	    notice(s_ChanServ, source, "  %3d %5d  %s%s%s%s",
			i+1, access->level, access->name,
			s ? " (" : "", s ? s : "", s ? ")" : "");
	}

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2ACCESS %s {ADD|DEL|LIST} [\37mask\37 [\37level\37]]\2", chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_ChanServ);

    }
}

/*************************************************************************/

static void do_akick(const char *source)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *s    = strtok(NULL, " ");
    char *t;
    ChannelInfo *ci;
    NickInfo *ni;
    User *u;
    int i;
    AutoKick *akick;
    Channel *c;
    struct c_userlist *u2;

    if (!cmd || (stricmp(cmd, "LIST") != 0 && !s)) {

	notice(s_ChanServ, source,
		"Syntax: \2AKICK %s {ADD|DEL|LIST} [\37nick-or-usermask\37]\2",
		chan ? chan : "\37channel\37");
	notice(s_ChanServ, source,
		"\2/msg %s HELP AKICK\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!(u = finduser(source)) || get_access(u, ci) < 10) {

	notice(s_ChanServ, source, "Access denied.");

    } else if (stricmp(cmd, "ADD") == 0) {

	NickInfo *ni = findnick(s);
	char *nick, *user, *host;

	if (!ni) {
	    split_usermask(s, &nick, &user, &host);
	    s = smalloc(strlen(nick)+strlen(user)+strlen(host)+3);
	    sprintf(s, "%s!%s@%s", nick, user, host);
	    free(nick);
	    free(user);
	    free(host);
	}
	for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i) {
	    if ((akick->is_nick ? stricmp(akick->name,s) :
						strcmp(akick->name,s)) == 0) {
		notice(s_ChanServ, source,
			"\2%s\2 already exists on %s AKICK list.",
			akick->name, chan);
		return;
	    }
	}
	++ci->akickcount;
	ci->akick = srealloc(ci->akick, sizeof(AutoKick) * ci->akickcount);
	akick = &ci->akick[ci->akickcount-1];
	akick->name = ni ? sstrdup(s) : s;
	akick->is_nick = (ni != NULL);
	akick->pad = 0;
	if (s = strtok(NULL, ""))
	    akick->reason = sstrdup(s);
	else
	    akick->reason = NULL;
	notice(s_ChanServ, source,
		"\2%s\2 added to %s AKICK list.", akick->name, chan);
	if (c = findchan(chan)) {
	    for (u2 = c->users; u2; u2 = u2->next) {
		if (ni ? stricmp(u2->user->nick, akick->name) :
			 match_usermask(akick->name, u2->user))
		    send_cmd(s_ChanServ, "KICK %s %s :%s",
				u2->user->nick, chan, s ? s :
				"You have been banned from this channel.");
	    }
	}

    } else if (stricmp(cmd, "DEL") == 0) {

	/* Special case: is it a number?  Only do search if it isn't. */
	if (strspn(s, "1234567890") == strlen(s) &&
				(i = atoi(s)) > 0 && i <= ci->akickcount) {
	    --i;
	    akick = &ci->akick[i];
	} else {
	    /* First try for an exact match; then, a case-insensitive one. */
	    for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i) {
		if (strcmp(akick->name, s) == 0)
		    break;
	    }
	    if (i == ci->akickcount) {
		for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i) {
		    if (stricmp(akick->name, s) == 0)
			break;
		}
	    }
	    if (i == ci->akickcount) {
		notice(s_ChanServ, source,
			"\2%s\2 not found on %s AKICK list.", s, chan);
		return;
	    }
	}
	notice(s_ChanServ, source,
		"\2%s\2 deleted from %s AKICK list.", akick->name, chan);
	free(akick->name);
	if (akick->reason)
	    free(akick->reason);
	--ci->akickcount;
	if (i < ci->akickcount)
	    bcopy(akick+1, akick, sizeof(AutoKick) * (ci->akickcount-i));
	if (ci->akickcount)
	    ci->akick = srealloc(ci->akick, sizeof(AutoKick) * ci->akickcount);
	else {
	    free(ci->akick);
	    ci->akick = NULL;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_ChanServ, source, "AKICK list for %s:", chan);
	for (akick = ci->akick, i = 0; i < ci->akickcount; ++akick, ++i) {
	    if (s && !match_wild(s, akick->name))
		continue;
	    if (ni = findnick(akick->name))
		t = ni->last_usermask;
	    else
		t = NULL;
	    notice(s_ChanServ, source, "  %3d %s%s%s%s%s%s%s",
			i+1, akick->name,
			t ? " (" : "", t ? t : "", t ? ")" : "",
			akick->reason ? " (" : "",
			akick->reason ? akick->reason : "",
			akick->reason ? ")" : "");
	}

    } else {

	notice(s_ChanServ, source,
		"Syntax: \2AKICK %s {ADD|DEL|LIST} [\37nick-or-usermask\37]\2",
		chan);
	notice(s_ChanServ, source,
		"\2/msg %s HELP AKICK\2 for more information.", s_ChanServ);

    }
}

/*************************************************************************/

static void do_info(const char *source)
{
    char *chan = strtok(NULL, " ");
    ChannelInfo *ci;
    NickInfo *ni;
    struct tm tm;
    char s[32], *t, *end;

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2INFO \37channel\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP INFO\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else {

	notice(s_ChanServ, source, "Information for channel %s:", chan);
	if (ni = findnick(ci->founder))
	    t = ni->last_usermask;
	else
	    t = NULL;
	notice(s_ChanServ, source,
		"    Founder    : %s%s%s%s",
			ci->founder, t ? " (" : "", t ? t : "", t ? ")" : "");
	notice(s_ChanServ, source,
		"    Description: %s", ci->desc);
	tm = *localtime(&ci->time_registered);
	notice(s_ChanServ, source,
		"    Registered : %s %2d %02d:%02d:%02d %d",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ci->last_used);
	notice(s_ChanServ, source,
		"    Last used  : %s %2d %02d:%02d:%02d %d",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ci->last_topic)
	    notice(s_ChanServ, source,
		"    Last topic : %s", ci->last_topic);
	end = s;
	*end = 0;
	if (ci->mlock_on || ci->mlock_key || ci->mlock_limit)
	    end += sprintf(end, "+%s%s%s%s%s%s%s%s",
				(ci->mlock_on & CMODE_I) ? "i" : "",
				(ci->mlock_key         ) ? "k" : "",
				(ci->mlock_limit       ) ? "l" : "",
				(ci->mlock_on & CMODE_M) ? "m" : "",
				(ci->mlock_on & CMODE_N) ? "n" : "",
				(ci->mlock_on & CMODE_P) ? "p" : "",
				(ci->mlock_on & CMODE_S) ? "s" : "",
				(ci->mlock_on & CMODE_T) ? "t" : "");
	if (ci->mlock_off)
	    end += sprintf(end, "-%s%s%s%s%s%s%s%s",
				(ci->mlock_off & CMODE_I) ? "i" : "",
				(ci->mlock_off & CMODE_K) ? "k" : "",
				(ci->mlock_off & CMODE_L) ? "l" : "",
				(ci->mlock_off & CMODE_M) ? "m" : "",
				(ci->mlock_off & CMODE_N) ? "n" : "",
				(ci->mlock_off & CMODE_P) ? "p" : "",
				(ci->mlock_off & CMODE_S) ? "s" : "",
				(ci->mlock_off & CMODE_T) ? "t" : "");
	notice(s_ChanServ, source,
		"    Mode lock  : %s", s);
    }
}

/*************************************************************************/

static void do_list(const char *source)
{
    char *chan = strtok(NULL, " ");
    ChannelInfo *ci;
    int nchans, i;
    char buf[512];

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2LIST \37pattern\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_ChanServ);

    } else {

	nchans = 0;
	notice(s_ChanServ, source, "List of entries matching \2%s\2:", chan);
	for (i = 33; i < 256; ++i) {
	    for (ci = chanlists[i]; ci; ci = ci->next) {
		if (ci->flags & CI_PRIVATE)
		    continue;
		if (strlen(ci->name)+strlen(ci->desc) > sizeof(buf))
		    continue;
		sprintf(buf, "%-20s  %s", ci->name, ci->desc);
		if (match_wild(chan, buf)) {
		    if (++nchans <= 50)
			notice(s_ChanServ, source, "    %s", buf);
		}
	    }
	}
	notice(s_ChanServ, source, "End of list - %d/%d matches shown.",
					nchans>50 ? 50 : nchans, nchans);
    }

}

/*************************************************************************/

static void do_invite(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2INVITE \37channel\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP INVITE\2 for more information.", s_ChanServ);

    } else if (!findchan(chan)) {

	notice(s_ChanServ, source, "Channel %s does not exist.", chan);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!u || get_access(u, ci) < 1) {

	notice(s_ChanServ, source, "Access denied.");

    } else {

	send_cmd(s_ChanServ, "INVITE %s %s", source, chan);

    }
}

/*************************************************************************/

static void do_unban(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;
    Channel *c;
    int i;
    char *av[3];

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2UNBAN \37channel\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP UNBAN\2 for more information.", s_ChanServ);

    } else if (!(c = findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s does not exist.", chan);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!u || get_access(u, ci) < 1) {

	notice(s_ChanServ, source, "Access denied.");

    } else {

	av[0] = chan;
	av[1] = sstrdup("+b");
	for (i = 0; i < c->bancount; ++i) {
	    if (match_usermask(c->bans[i], u)) {
		send_cmd(s_ChanServ, "MODE %s -b %s", chan, c->bans[i]);
		av[2] = sstrdup(c->bans[i]);
		do_cmode(s_ChanServ, 3, av);
		free(av[2]);
	    }
	}
	free(av[1]);
	notice(s_ChanServ, source, "You have been unbanned from %s.", chan);

    }
}

/*************************************************************************/

static void do_getpass(const char *source)
{
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    ChannelInfo *ci;

    if (!chan) {

	notice(s_ChanServ, source, "Syntax: \2GETPASS \37channel\37\2");
	notice(s_ChanServ, source,
		"\2/msg %s HELP GETPASS\2 for more information.", s_ChanServ);

    } else if (!(ci = cs_findchan(chan))) {

	notice(s_ChanServ, source, "Channel %s is not registered.", chan);

    } else if (!u) {

	notice(s_ChanServ, source, "Could not find your user record!");

    } else {

	log("%s: %s!%s@%s used GETPASS on %s",
		s_ChanServ, source, u->username, u->host, chan);
	wallops("%s used GETPASS command on %s", source, chan);
	notice(s_ChanServ, source, "Password for %s is: \2%s\2",
		chan, ci->founderpass);

    }
}
