/* Main services code */

#include "services.h"


/* Set to 1 if we are waiting for input */
static int waiting = 0;

/* Set to 1 after we've set everything up */
static int started = 0;

/* Set to 1 if we are to quit */
int quitting = 0;

/* Contains a message as to why services is terminating */
char *quitmsg = NULL;

/* Input buffer - global, so we can dump it if something goes wrong */
char inbuf[BUFSIZE];

/* Socket for talking to server */
int servsock;

/* Should we update the databases now? */
int save_data = 0;

/* If we get a signal, use this to jump out of the main loop. */
static jmp_buf panic_jmp;

/*************************************************************************/

/* If we get a weird signal, come here. */

static void sighandler(int signum)
{
    if (started) {
	if (!waiting) {
	    log("PANIC! buffer = %s", inbuf);
	    /* Cut off if this would make IRC command >510 characters. */
	    if (strlen(inbuf) > 448) {
		inbuf[446] = '>';
		inbuf[447] = '>';
		inbuf[448] = 0;
	    }
	    wallops("PANIC! buffer = %s\r\n", inbuf);
	} else if (waiting < 0) {
	    if (waiting == -1) {
		log("PANIC! in timed_update");
		wallops("PANIC! in timed_update");
	    } else {
		log("PANIC! waiting=%d", waiting);
		wallops("PANIC! waiting=%d", waiting);
	    }
	}
    }
    if (signum == SIGUSR1 || !(quitmsg = malloc(256))) {
	quitmsg = "Out of memory!";
	quitting = 1;
    } else {
#ifdef HAS_STRSIGNAL
	sprintf(quitmsg, "Services terminating: %s", strsignal(signum));
#else
	sprintf(quitmsg, "Services terminating on signal %d", signum);
#endif
	quitting = 1;
    }
    if (started)
	longjmp(panic_jmp, 1);
    else {
	log("%s", quitmsg);
	exit(1);
    }
}

/*************************************************************************/

/* Log stuff to stderr with a datestamp. */

void log(const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
    strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
    fputs(buf, stderr);
    vfprintf(stderr, fmt, args);
    fputc('\n', stderr);
    fflush(stderr);
}

void log_perror(const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[256];

    va_start(args, fmt);
    time(&t);
    tm = *localtime(&t);
    strftime(buf, sizeof(buf)-1, "[%b %d %H:%M:%S %Y] ", &tm);
    fputs(buf, stderr);
    vfprintf(stderr, fmt, args);
    fprintf(stderr, ": %s\n", strerror(errno));
    fflush(stderr);
}

/*************************************************************************/

/* Main routine.  (What does it look like? :-) ) */

int main(int ac, char **av)
{
    time_t last_update;	/* When did we last update the databases? */
    time_t last_check;	/* When did we last check NickServ timeouts? */
    int i;
    char *progname;


    /* Find program name. */
    if (progname = strrchr(av[0], '/'))
	++progname;
    else
	progname = av[0];


    /* Chdir to services directory. */

    if (chdir(SERVICES_DIR) < 0) {
	perror(SERVICES_DIR);
	return 20;
    }


    /* Check for program name == "listnicks" or "listchans" and do
     * appropriate special stuff if so. */

    if (strcmp(progname, "listnicks") == 0) {

	int count = 0;	/* Count only rather than display? */
	int usage = 0;	/* Display command usage?  (>0 also indicates error) */
	int i;

	i = 1;
	while (i < ac) {
	    if (av[i][0] == '-') {
		switch (av[i][1]) {
		case 'h':
		    usage = -1; break;
		case 'c':
		    if (i > 1)
			usage = 1;
		    count =  1; break;
		default :
		    usage =  1; break;
		}
		--ac;
		if (i < ac)
		bcopy(av+i+1, av+i, sizeof(char *) * ac-i);
	    } else {
		if (count)
		    usage = 1;
		++i;
	    }
	}
	if (usage) {
	    fprintf(stderr, "\
\n\
Usage: listnicks [-c] [nick [nick...]]\n\
     -c: display count of registered nicks only\n\
            (cannot be combined with nicks)\n\
   nick: nickname(s) to display information for\n\
\n\
If no nicks are given, the entire nickname database is printed out in\n\
compact format followed by the number of registered nicks (with -c, the\n\
list is suppressed and only the count is printed).  If one or more nicks\n\
are given, detailed information about those nicks is displayed.\n\
\n");
	    return usage>0 ? 1 : 0;
	}
	load_ns_dbase();
	if (ac > 1) {
	    for (i = 1; i < ac; ++i)
		listnicks(0, av[i]);
	} else {
	    listnicks(count, NULL);
	}
	return 0;

    } else if (strcmp(progname, "listchans") == 0) {

	int count = 0;	/* Count only rather than display? */
	int usage = 0;	/* Display command usage?  (>0 also indicates error) */
	int i;

	i = 1;
	while (i < ac) {
	    if (av[i][0] == '-') {
		switch (av[i][1]) {
		case 'h':
		    usage = -1; break;
		case 'c':
		    if (i > 1)
			usage = 1;
		    count =  1; break;
		default :
		    usage =  1; break;
		}
		--ac;
		if (i < ac)
		bcopy(av+i+1, av+i, sizeof(char *) * ac-i);
	    } else {
		if (count)
		    usage = 1;
		++i;
	    }
	}
	if (usage) {
	    fprintf(stderr, "\
\n\
Usage: listchans [-c] [chan [chan...]]\n\
     -c: display count of registered channels only\n\
            (cannot be combined with channels)\n\
   chan: channel(s) to display information for\n\
\n\
If no channels are given, the entire channel database is printed out in\n\
compact format followed by the number of registered channelss (with -c, the\n\
list is suppressed and only the count is printed).  If one or more channels\n\
are given, detailed information about those channels is displayed.\n\
\n");
	    return usage>0 ? 1 : 0;
	}
	load_ns_dbase();
	load_cs_dbase();
	if (ac > 1) {
	    for (i = 1; i < ac; ++i)
		listchans(0, av[i]);
	} else {
	    listchans(count, NULL);
	}
	return 0;

    }


    /* Redirect stderr to logfile. */

    if (!freopen(LOG_FILENAME, "a", stderr)) {
	perror(LOG_FILENAME);
	return 20;
    }


    /* Detach ourselves */

    if ((i = fork()) < 0) {
	log_perror("fork()");
	return 1;
    } else if (i != 0)
	return 0;
    if (setpgrp(0, 0) < 0) {
	log_perror("setpgrp()");
	return 1;
    }


    log("Services starting up");


    /* Set signal handlers */

    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);
    signal(SIGPIPE, SIG_IGN);		/* We don't care about broken pipes */
    signal(SIGQUIT, sighandler);
    signal(SIGSEGV, sighandler);
    signal(SIGBUS, sighandler);
    signal(SIGQUIT, sighandler);
    signal(SIGHUP, sighandler);
    signal(SIGILL, sighandler);
    signal(SIGTRAP, sighandler);
    signal(SIGIOT, sighandler);
    signal(SIGFPE, sighandler);
    signal(SIGALRM, SIG_IGN);		/* Used by sgets() for read timeout */
    signal(SIGUSR2, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);

    signal(SIGUSR1, sighandler);	/* This is our "out-of-memory" panic switch */


    /* Load up databases. */

    load_ns_dbase();
    load_cs_dbase();
    load_akill();


    /* Connect to the remote server */

    servsock = conn(REMOTE_SERVER, REMOTE_PORT);
    if (servsock < 0) {
	log_perror("Can't connect to server");
	return 20;
    }
    send_cmd(NULL, "PASS :%s\r\n", PASSWORD);
    send_cmd(NULL, "SERVER %s 1 :%s\r\n", SERVER_NAME, SERVER_DESC);
    sgets2(inbuf, sizeof(inbuf), servsock);
    if (strnicmp(inbuf, "ERROR", 5) == 0) {
	log("Remote server returned: %s", inbuf);
	return 20;
    }
    send_cmd(NULL, "NICK NickServ 1 %lu %s %s %s :Nickname Server\r\n",
			time(NULL), SERVICES_USER, SERVICES_HOST, SERVER_NAME);
    send_cmd(NULL, "NICK ChanServ 1 %lu %s %s %s :Channel Server\r\n",
			time(NULL), SERVICES_USER, SERVICES_HOST, SERVER_NAME);
    send_cmd(NULL, "NICK HelpServ 1 %lu %s %s %s :Help Server\r\n",
			time(NULL), SERVICES_USER, SERVICES_HOST, SERVER_NAME);
    send_cmd(NULL, "NICK MemoServ 1 %lu %s %s %s :Memo Server\r\n",
			time(NULL), SERVICES_USER, SERVICES_HOST, SERVER_NAME);
    send_cmd(NULL, "NICK OperServ 1 %lu %s %s %s :Operator Server\r\n",
			time(NULL), SERVICES_USER, SERVICES_HOST, SERVER_NAME);
    send_cmd(s_OperServ, "MODE OperServ +i\r\n");


    /* Main loop. */

    /* We have a line left over from earlier, so process it first. */
    process();

    /* The signal handler routine will drop back here with quitting != 0
     * if it gets called. */
    setjmp(panic_jmp);
    started = 1;

    last_update = time(NULL);

    while (!quitting) {
	time_t t = time(NULL);
	waiting = -3;
	if (save_data || t-last_update >= UPDATE_TIMEOUT) {
	    waiting = -22;
	    /* First check for expired nicks/channels */
	    expire_nicks();
	    expire_chans();
	    /* Now actually save stuff */
	    waiting = -2;
	    save_ns_dbase();
	    save_cs_dbase();
	    save_akill();
	    save_data = 0;
	    last_update = t;
	}
	waiting = -1;
	/* Ew... hardcoded constant.  Somebody want to come up with a good
	 * name for this? */
	if (t-last_check >= 3)
	    check_timeouts();
	waiting = 1;
	i = (int)sgets2(inbuf, sizeof(inbuf), servsock);
	waiting = 0;
	if (i > 0) {
	    process();
	} else if (i == 0) {
	    if (quitmsg = malloc(512))
		sprintf(quitmsg, "Read error from server: %s", strerror(errno));
	    else
		quitmsg = "Read error from server";
	    quitting = 1;
	}
	waiting = -4;
    }


    /* Disconnect and exit */

    if (!quitmsg)
	quitmsg = "Terminating, reason unknown";
    log("%s", quitmsg);
    if (started)
	send_cmd(SERVER_NAME, "SQUIT %s :%s\r\n", SERVER_NAME, quitmsg);
    disconn(servsock);
    return 0;
}
