/* Miscellaneous functions */

#include "services.h"

/*************************************************************************/

/* sgets(), but clean off newline-type stuff. */
char *sgets2(char *buf, long size, int sock)
{
    char *s = sgets(buf, size, sock);
    if (!s || s == (char *)-1)
	return s;
    if (buf[strlen(buf)-1] == '\n')
	buf[strlen(buf)-1] = 0;
    if (buf[strlen(buf)-1] == '\r')
	buf[strlen(buf)-1] = 0;
    return buf;
}

/*************************************************************************/

/* Like strncpy(), but always adds a null character. */
char *strscpy(char *d, const char *s, int len)
{
    char *d_orig = d;

    if (!len)
	return d;
    while (--len && (*d++ = *s++))
	;
    *d = 0;
    return d_orig;
}

/*************************************************************************/

#ifdef NEED_STRICMP
int stricmp(const char *s1, const char *s2)
{
    register char c;

    while ((c = tolower(*s1)) == tolower(*s2)) {
	if (c == 0) return 0;
	s1++; s2++;
    }
    if (c < tolower(*s2))
	return -1;
    return 1;
}

int strnicmp(const char *s1, const char *s2, int len)
{
    register char c;

    if (!len) return 0;
    while ((c = tolower(*s1)) == tolower(*s2) && len > 0) {
	if (c == 0 || --len == 0) return 0;
	s1++; s2++;
    }
    if (c < tolower(*s2))
	return -1;
    return 1;
}
#endif

/*************************************************************************/

#ifdef NEED_STRDUP
char *strdup(const char *s)
{
    char *new = malloc(strlen(s)+1);
    if (new)
	strcpy(new, s);
    return new;
}
#endif

/*************************************************************************/

char *stristr(char *s1, char *s2)
{
    register char *s = s1, *d = s2;

    while (*s1) {
	if (tolower(*s1) == tolower(*d)) {
	    s1++; d++;
	    if (*d == 0)
		return s;
	} else {
	    s = ++s1;
	    d = s2;
	}
    }
    return NULL;
}

/*************************************************************************/

#ifdef NEED_STRERROR
extern char *__sys_errlist[];
char *strerror(int errnum)
{
    return __sys_errlist[errnum];
}
#endif

/*************************************************************************/

/* This is a rather ugly kludge to get strtok() to do what we want; i.e.
 * return a parameter that's not followed by a delimiter character.  This
 * strtok() will only return NULL if either (1) strtok(NULL,...) is called
 * before strtok(str,...), or (2) no characters are left in the string. */

char *strtok(char *s, const char *delim)
{
    static char *pos = NULL;
    char *res;

    char *t;
    if (s)
	pos = s;
    if (!pos || !*pos)
	return NULL;
    res = pos;
    pos = strpbrk(pos, delim);
    if (pos && *pos)
	*pos++ = 0;
    return res;
}

/*************************************************************************/

/* Memory allocation: if we run out of memory, quit. */

void *smalloc(long size)
{
    void *buf;

    buf = malloc(size);
    if (!size)
	log("smalloc: Illegal attempt to allocate 0 bytes");
    if (!buf)
	raise(SIGUSR1);
    return buf;
}

void *scalloc(long elsize, long els)
{
    void *buf = calloc(elsize, els);
    if (!elsize || !els)
	log("scalloc: Illegal attempt to allocate 0 bytes");
    if (!buf)
	raise(SIGUSR1);
    return buf;
}

void *srealloc(void *oldptr, long newsize)
{
    void *buf = realloc(oldptr, newsize);
    if (!newsize)
	log("srealloc: Illegal attempt to allocate 0 bytes");
    if (!buf)
	raise(SIGUSR1);
    return buf;
}

char *sstrdup(const char *s)
{
    char *t = strdup(s);
    if (!t)
	raise(SIGUSR1);
    return t;
}

/*************************************************************************/

/* Create a single big string out of an argc/argv. */
char *merge_args(int argc, char **argv)
{
    int i;
    static char s[4096];
    char *t;

    t = s;
    for (i = 0; i < argc; ++i)
	t += sprintf(t, "%s%s", *argv++, (i<argc-1) ? " " : "");
    return s;
}

/*************************************************************************/

/* See if the string matches the wildcard pattern. */
int match_wild(const char *pattern, const char *str)
{
    char c;
    const char *s;

    /* This WILL eventually terminate: either by *pattern == 0, or by a
     * trailing '*'. */

    for (;;) {
	switch (c = *pattern++) {
	case 0:
	    if (!*str)
		return 1;
	    return 0;
	case '?':
	    ++str;
	    break;
	case '*':
	    if (!*pattern)
		return 1;	/* trailing '*' matches everything else */
	    s = str;
	    while (*s) {
		if (*s == *pattern && match_wild(pattern, s))
		    return 1;
		++s;
	    }
	    break;
	default:
	    if (*str++ != c)
		return 0;
	    break;
	} /* switch */
    }
}

/*************************************************************************/

/* Return the three-character month name corresponding to the given month
 * number, 1-12. */

static char months[12][4] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

char *month_name(int month)
{
    return months[month-1];
}

/*************************************************************************/

/* Convert a string to upper/lower case. */

char *strupper(char *s)
{
    char *t = s;
    while (*t)
	*t++ = toupper(*t);
    return s;
}

char *strlower(char *s)
{
    char *t = s;
    while (*t)
	*t++ = tolower(*t);
    return s;
}

/*************************************************************************/

/* Read a string from a file, or write a string to a file, with the string
 * length prefixed as a two-byte big-endian integer.  Filename is passed in
 * in case of errors. */

char *read_string(FILE *f, const char *filename)
{
    char *s;
    int len;

    len = fgetc(f) * 256 + fgetc(f);
    s = smalloc(len);
    if (len != fread(s, 1, len, f)) {
	log_perror("Read error on %s", filename);
	exit(1);
    }
    return s;
}

char *write_string(const char *s, FILE *f, const char *filename)
{
    int len;

    len = strlen(s) + 1;	/* Include trailing null */
    fputc(len / 256, f); fputc(len & 255, f);
    if (len != fwrite(s, 1, len, f)) {
	log_perror("Write error on %s", filename);
	exit(1);
    }
}
